# -*- coding: utf-8 -*-
import ast
import json
import math
import os
import re
import secrets
import string
import time as timenow
from datetime import date, datetime, time, timedelta, timezone

from ruamel.yaml import YAML

from nectargraphenebase.account import PasswordKey

timeFormat = "%Y-%m-%dT%H:%M:%S"
# https://github.com/matiasb/python-unidiff/blob/master/unidiff/constants.py#L37
# @@ (source offset, length) (target offset, length) @@ (section header)
RE_HUNK_HEADER = re.compile(
    r"^@@ -(\d+)(?:,(\d+))? \+(\d+)(?:,(\d+))?\ @@[ ]?(.*)$", flags=re.MULTILINE
)


def formatTime(t):
    """Properly Format Time for permlinks"""
    if isinstance(t, float):
        return datetime.fromtimestamp(t, tz=timezone.utc).strftime("%Y%m%dt%H%M%S%Z")
    if isinstance(t, (datetime, date, time)):
        return t.strftime("%Y%m%dt%H%M%S%Z")


def addTzInfo(t, timezone_str="UTC"):
    """Returns a datetime object with tzinfo added
    Uses Python's built-in timezone when possible
    """
    if t and isinstance(t, (datetime, date, time)) and t.tzinfo is None:
        # Use built-in timezone
        if timezone_str.upper() == "UTC":
            if isinstance(t, datetime):
                t = t.replace(tzinfo=timezone.utc)
            # For date objects that don't have tzinfo directly
            elif isinstance(t, date) and not isinstance(t, datetime):
                t = datetime.combine(t, time.min).replace(tzinfo=timezone.utc)
            elif isinstance(t, time):
                t = datetime.combine(date.today(), t).replace(tzinfo=timezone.utc)
        else:
            # For non-UTC timezones, we can't use pytz anymore
            # This is a simplified approach - in the future, consider using zoneinfo for Python 3.9+
            # For now, default to UTC with a warning
            if isinstance(t, datetime):
                t = t.replace(tzinfo=timezone.utc)
            elif isinstance(t, date) and not isinstance(t, datetime):
                t = datetime.combine(t, time.min).replace(tzinfo=timezone.utc)
            elif isinstance(t, time):
                t = datetime.combine(date.today(), t).replace(tzinfo=timezone.utc)
            print(
                f"Warning: Non-UTC timezone '{timezone_str}' not supported without pytz. Using UTC instead."
            )
    return t


def formatTimeString(t):
    """Properly Format Time for permlinks"""
    if isinstance(t, (datetime, date, time)):
        return t.strftime(timeFormat)
    return addTzInfo(datetime.strptime(t, timeFormat))


def formatToTimeStamp(t):
    """Returns a timestamp integer

    :param datetime t: datetime object
    :return: Timestamp as integer
    """
    if isinstance(t, (datetime, date, time)):
        t = addTzInfo(t)
    else:
        t = formatTimeString(t)
    epoch = addTzInfo(datetime(1970, 1, 1))
    return int((t - epoch).total_seconds())


def formatTimeFromNow(secs=0):
    """Properly Format Time that is `x` seconds in the future

    :param int secs: Seconds to go in the future (`x>0`) or the
                     past (`x<0`)
    :return: Properly formated time for Graphene (`%Y-%m-%dT%H:%M:%S`)
    :rtype: str

    """
    return datetime.fromtimestamp(timenow.time() + int(secs), tz=timezone.utc).strftime(timeFormat)


def formatTimedelta(td):
    """Format timedelta to String"""
    if not isinstance(td, timedelta):
        return ""
    days, seconds = td.days, td.seconds
    hours = days * 24 + seconds // 3600
    minutes = (seconds % 3600) // 60
    seconds = seconds % 60
    return "%d:%s:%s" % (hours, str(minutes).zfill(2), str(seconds).zfill(2))


def parse_time(block_time):
    """Take a string representation of time from the blockchain, and parse it
    into datetime object.
    """
    return datetime.strptime(block_time, timeFormat).replace(tzinfo=timezone.utc)


def assets_from_string(text):
    """Correctly split a string containing an asset pair.

    Splits the string into two assets with the separator being on of the
    following: ``:``, ``/``, or ``-``.
    """
    return re.split(r"[\-:\/]", text)


def sanitize_permlink(permlink):
    permlink = permlink.strip()
    permlink = re.sub(r"_|\s|\.", "-", permlink)
    permlink = re.sub(r"[^\w-]", "", permlink)
    permlink = re.sub(r"[^a-zA-Z0-9-]", "", permlink)
    permlink = permlink.lower()
    return permlink


def derive_permlink(
    title, parent_permlink=None, parent_author=None, max_permlink_length=256, with_suffix=True
):
    """Derive a permlink from a comment title (for root level
    comments) or the parent permlink and optionally the parent
    author (for replies).

    """
    suffix = "-" + formatTime(datetime.now(timezone.utc)).lower()
    if parent_permlink and parent_author:
        prefix = "re-" + sanitize_permlink(parent_author) + "-"
        if with_suffix:
            rem_chars = max_permlink_length - len(suffix) - len(prefix)
        else:
            rem_chars = max_permlink_length - len(prefix)
        body = sanitize_permlink(parent_permlink)[:rem_chars]
        if with_suffix:
            return prefix + body + suffix
        else:
            return prefix + body
    elif parent_permlink:
        prefix = "re-"
        if with_suffix:
            rem_chars = max_permlink_length - len(suffix) - len(prefix)
        else:
            rem_chars = max_permlink_length - len(prefix)
        body = sanitize_permlink(parent_permlink)[:rem_chars]
        if with_suffix:
            return prefix + body + suffix
        else:
            return prefix + body
    else:
        if with_suffix:
            rem_chars = max_permlink_length - len(suffix)
        else:
            rem_chars = max_permlink_length
        body = sanitize_permlink(title)[:rem_chars]
        if len(body) == 0:  # empty title or title consisted of only special chars
            return suffix[1:]  # use timestamp only, strip leading "-"
        if with_suffix:
            return body + suffix
        else:
            return body


def resolve_authorperm(identifier):
    """Correctly split a string containing an authorperm.

    Splits the string into author and permlink with the
    following separator: ``/``.

    Examples:

        .. code-block:: python

            >>> from nectar.utils import resolve_authorperm
            >>> author, permlink = resolve_authorperm('https://d.tube/#!/v/pottlund/m5cqkd1a')
            >>> author, permlink = resolve_authorperm("https://steemit.com/witness-category/@gtg/24lfrm-gtg-witness-log")
            >>> author, permlink = resolve_authorperm("@gtg/24lfrm-gtg-witness-log")
            >>> author, permlink = resolve_authorperm("https://busy.org/@gtg/24lfrm-gtg-witness-log")

    """
    # without any http(s)
    match = re.match(r"@?([\w\-\.]*)/([\w\-]*)", identifier)
    if hasattr(match, "group"):
        return match.group(1), match.group(2)
    # dtube url
    match = re.match(r"([\w\-\.]+[^#?\s]+)/#!/v/?([\w\-\.]*)/([\w\-]*)", identifier)
    if hasattr(match, "group"):
        return match.group(2), match.group(3)
    # url
    match = re.match(r"([\w\-\.]+[^#?\s]+)/@?([\w\-\.]*)/([\w\-]*)", identifier)
    if not hasattr(match, "group"):
        raise ValueError("Invalid identifier")
    return match.group(2), match.group(3)


def construct_authorperm(*args):
    """Create a post identifier from comment/post object or arguments.
    Examples:

        .. code-block:: python

            >>> from nectar.utils import construct_authorperm
            >>> print(construct_authorperm('username', 'permlink'))
            @username/permlink
            >>> print(construct_authorperm({'author': 'username', 'permlink': 'permlink'}))
            @username/permlink

    """
    username_prefix = "@"
    if len(args) == 1:
        op = args[0]
        author, permlink = op["author"], op["permlink"]
    elif len(args) == 2:
        author, permlink = args
    else:
        raise ValueError("construct_identifier() received unparsable arguments")

    fields = dict(prefix=username_prefix, author=author, permlink=permlink)
    return "{prefix}{author}/{permlink}".format(**fields)


def resolve_root_identifier(url):
    m = re.match(r"/([^/]*)/@([^/]*)/([^#]*).*", url)
    if not m:
        return "", ""
    else:
        category = m.group(1)
        author = m.group(2)
        permlink = m.group(3)
        return construct_authorperm(author, permlink), category


def resolve_authorpermvoter(identifier):
    """Correctly split a string containing an authorpermvoter.

    Splits the string into author and permlink with the
    following separator: ``/`` and ``|``.
    """
    pos = identifier.find("|")
    if pos < 0:
        raise ValueError("Invalid identifier")
    [author, permlink] = resolve_authorperm(identifier[:pos])
    return author, permlink, identifier[pos + 1 :]


def construct_authorpermvoter(*args):
    """Create a vote identifier from vote object or arguments.
    Examples:

        .. code-block:: python

            >>> from nectar.utils import construct_authorpermvoter
            >>> print(construct_authorpermvoter('username', 'permlink', 'voter'))
            @username/permlink|voter
            >>> print(construct_authorpermvoter({'author': 'username', 'permlink': 'permlink', 'voter': 'voter'}))
            @username/permlink|voter

    """
    username_prefix = "@"
    if len(args) == 1:
        op = args[0]
        if "authorperm" in op:
            authorperm, voter = op["authorperm"], op["voter"]
            [author, permlink] = resolve_authorperm(authorperm)
        else:
            author, permlink, voter = op["author"], op["permlink"], op["voter"]
    elif len(args) == 2:
        authorperm, voter = args
        [author, permlink] = resolve_authorperm(authorperm)
    elif len(args) == 3:
        author, permlink, voter = args
    else:
        raise ValueError("construct_identifier() received unparsable arguments")

    fields = dict(prefix=username_prefix, author=author, permlink=permlink, voter=voter)
    return "{prefix}{author}/{permlink}|{voter}".format(**fields)


def reputation_to_score(rep):
    """Converts the account reputation value into the reputation score"""
    if isinstance(rep, str):
        rep = int(rep)
    if rep == 0:
        return 25.0
    score = max([math.log10(abs(rep)) - 9, 0])
    if rep < 0:
        score *= -1
    score = (score * 9.0) + 25.0
    return score


def remove_from_dict(obj, keys=list(), keep_keys=True):
    """Prune a class or dictionary of all but keys (keep_keys=True).
    Prune a class or dictionary of specified keys.(keep_keys=False).
    """
    if not isinstance(obj, dict):
        obj = dict(obj)
    if keep_keys:
        return {k: v for k, v in obj.items() if k in keys}
    else:
        return {k: v for k, v in obj.items() if k not in keys}


def make_patch(a, b):
    import diff_match_patch as dmp_module

    dmp = dmp_module.diff_match_patch()
    patch = dmp.patch_make(a, b)
    patch_text = dmp.patch_toText(patch)
    return patch_text


def findall_patch_hunks(body=None):
    return RE_HUNK_HEADER.findall(body)


def derive_beneficiaries(beneficiaries):
    beneficiaries_list = []
    beneficiaries_accounts = []
    beneficiaries_sum = 0
    if not isinstance(beneficiaries, list):
        beneficiaries = beneficiaries.split(",")

    for w in beneficiaries:
        account_name = w.strip().split(":")[0]
        if account_name[0] == "@":
            account_name = account_name[1:]
        if account_name in beneficiaries_accounts:
            continue
        if w.find(":") == -1:
            percentage = -1
        else:
            percentage = w.strip().split(":")[1]
            if "%" in percentage:
                percentage = percentage.strip().split("%")[0].strip()
            percentage = float(percentage)
            beneficiaries_sum += percentage
        beneficiaries_list.append({"account": account_name, "weight": int(percentage * 100)})
        beneficiaries_accounts.append(account_name)

    missing = 0
    for bene in beneficiaries_list:
        if bene["weight"] < 0:
            missing += 1
    index = 0
    for bene in beneficiaries_list:
        if bene["weight"] < 0:
            beneficiaries_list[index]["weight"] = int(
                (int(100 * 100) - int(beneficiaries_sum * 100)) / missing
            )
        index += 1
    sorted_beneficiaries = sorted(
        beneficiaries_list, key=lambda beneficiaries_list: beneficiaries_list["account"]
    )
    return sorted_beneficiaries


def derive_tags(tags):
    tags_list = []
    if len(tags.split(",")) > 1:
        for tag in tags.split(","):
            tags_list.append(tag.strip())
    elif len(tags.split(" ")) > 1:
        for tag in tags.split(" "):
            tags_list.append(tag.strip())
    elif len(tags) > 0:
        tags_list.append(tags.strip())
    return tags_list


def seperate_yaml_dict_from_body(content):
    parameter = {}
    body = ""
    if len(content.split("---\n")) > 1:
        body = content[content.find("---\n", 1) + 4 :]
        yaml_content = content[content.find("---\n") + 4 : content.find("---\n", 1)]
        yaml = YAML(typ="safe")
        parameter = yaml.load(yaml_content)
        if not isinstance(parameter, dict):
            parameter = yaml.load(yaml_content.replace(":", ": ").replace("  ", " "))
    else:
        body = content
    return body, parameter


def create_yaml_header(comment, json_metadata={}, reply_identifier=None):
    yaml_prefix = "---\n"
    if comment["title"] != "":
        yaml_prefix += 'title: "%s"\n' % comment["title"]
    if "permlink" in comment:
        yaml_prefix += "permlink: %s\n" % comment["permlink"]
    yaml_prefix += "author: %s\n" % comment["author"]
    if "author" in json_metadata:
        yaml_prefix += "authored by: %s\n" % json_metadata["author"]
    if "description" in json_metadata:
        yaml_prefix += 'description: "%s"\n' % json_metadata["description"]
    if "canonical_url" in json_metadata:
        yaml_prefix += "canonical_url: %s\n" % json_metadata["canonical_url"]
    if "app" in json_metadata:
        yaml_prefix += "app: %s\n" % json_metadata["app"]
    if "last_update" in comment:
        yaml_prefix += "last_update: %s\n" % comment["last_update"]
    elif "updated" in comment:
        yaml_prefix += "last_update: %s\n" % comment["updated"]
    yaml_prefix += "max_accepted_payout: %s\n" % str(comment["max_accepted_payout"])
    if "percent_steem_dollars" in comment:
        yaml_prefix += "percent_steem_dollars: %s\n" % str(comment["percent_steem_dollars"])
    elif "percent_hbd" in comment:
        yaml_prefix += "percent_hbd: %s\n" % str(comment["percent_hbd"])
    if "tags" in json_metadata:
        if (
            len(json_metadata["tags"]) > 0
            and comment["category"] != json_metadata["tags"][0]
            and len(comment["category"]) > 0
        ):
            yaml_prefix += "community: %s\n" % comment["category"]
        yaml_prefix += "tags: %s\n" % ",".join(json_metadata["tags"])
    if "beneficiaries" in comment:
        beneficiaries = []
        for b in comment["beneficiaries"]:
            beneficiaries.append("%s:%.2f%%" % (b["account"], b["weight"] / 10000 * 100))
        if len(beneficiaries) > 0:
            yaml_prefix += "beneficiaries: %s\n" % ",".join(beneficiaries)
    if reply_identifier is not None:
        yaml_prefix += "reply_identifier: %s\n" % reply_identifier
    yaml_prefix += "---\n"
    return yaml_prefix


def load_dirty_json(dirty_json):
    regex_replace = [
        (r"([ \{,:\[])(u)?'([^']+)'", r'\1"\3"'),
        (r" False([, \}\]])", r" false\1"),
        (r" True([, \}\]])", r" true\1"),
    ]
    for r, s in regex_replace:
        dirty_json = re.sub(r, s, dirty_json)
    clean_json = json.loads(dirty_json)
    return clean_json


def create_new_password(length=32):
    """Creates a random password containing alphanumeric chars with at least 1 number and 1 upper and lower char"""
    alphabet = string.ascii_letters + string.digits
    while True:
        import_password = "".join(secrets.choice(alphabet) for i in range(length))
        if (
            any(c.islower() for c in import_password)
            and any(c.isupper() for c in import_password)
            and any(c.isdigit() for c in import_password)
        ):
            break
    return import_password


def import_coldcard_wif(filename):
    """Reads a exported coldcard Wif text file and returns the WIF and used path"""
    next_var = ""
    import_password = ""
    path = ""
    with open(filename) as fp:
        for line in fp:
            if line.strip() == "":
                continue
            if line.strip() == "WIF (privkey):":
                next_var = "wif"
                continue
            elif "Path Used" in line.strip():
                next_var = "path"
                continue
            if next_var == "wif":
                import_password = line.strip()
            elif next_var == "path":
                path = line
            next_var = ""
    return import_password, path.lstrip().replace("\n", "")


def generate_password(import_password, wif=1):
    if wif > 0:
        password = import_password
        for _ in range(wif):
            pk = PasswordKey("", password, role="")
            password = str(pk.get_private())
        password = "P" + password
    else:
        password = import_password
    return password


def import_pubkeys(import_pub):
    if not os.path.isfile(import_pub):
        raise Exception("File %s does not exist!" % import_pub)
    with open(import_pub) as fp:
        pubkeys = fp.read()
    if pubkeys.find("\0") > 0:
        with open(import_pub, encoding="utf-16") as fp:
            pubkeys = fp.read()
    pubkeys = ast.literal_eval(pubkeys)
    owner = pubkeys["owner"]
    active = pubkeys["active"]
    posting = pubkeys["posting"]
    memo = pubkeys["memo"]
    return owner, active, posting, memo


def import_custom_json(jsonid, json_data):
    """Returns a list of required authorities for a custom_json operation.

    Returns the author and required posting authorities for a custom_json operation.

    Args:
        jsonid: The id of the custom json
        json_data: The data of the custom json

    Returns:
        tuple with required author and posting authorities
    """
    try:
        if (
            isinstance(json_data, dict)
            and "required_auths" in json_data
            and "required_posting_auths" in json_data
        ):
            required_auths = json_data["required_auths"]
            required_posting_auths = json_data["required_posting_auths"]
            del json_data["required_auths"]
            del json_data["required_posting_auths"]
            return required_auths, required_posting_auths
        else:
            return [], []
    except (KeyError, ValueError, TypeError):
        return [], []



"""nectarbase."""

from .version import version as __version__

__all__ = [
    "__version__",
    "memo",
    "objects",
    "objecttypes",
    "operationids",
    "operations",
    "signedtransactions",
    "ledgertransactions",
    "transactions",
]



import os
import unittest

from nectarstorage.base import InRamConfigurationStore, InRamEncryptedKeyStore
from nectarstorage.exceptions import WrongMasterPasswordException


class Testcases(unittest.TestCase):
    def test_masterpassword(self):
        password = "foobar"
        config = InRamConfigurationStore()
        keys = InRamEncryptedKeyStore(config=config)
        self.assertFalse(keys.has_masterpassword())
        master = keys._new_masterpassword(password)
        self.assertEqual(
            len(master),
            len("66eaab244153031e8172e6ffed3217288515ddb63646bbefa981a654bdf25b9f"),
        )
        with self.assertRaises(Exception):
            keys._new_masterpassword(master)

        keys.lock()

        with self.assertRaises(Exception):
            keys.change_password("foobar")

        keys.unlock(password)
        self.assertEqual(keys.decrypted_master, master)

        new_pass = "new_secret_password"
        keys.change_password(new_pass)
        keys.lock()
        keys.unlock(new_pass)
        self.assertEqual(keys.decrypted_master, master)

    def test_wrongmastermass(self):
        config = InRamConfigurationStore()
        keys = InRamEncryptedKeyStore(config=config)
        keys._new_masterpassword("foobar")
        keys.lock()
        with self.assertRaises(WrongMasterPasswordException):
            keys.unlock("foobar2")

    def test_masterpwd(self):
        with self.assertRaises(Exception):
            InRamEncryptedKeyStore()
        config = InRamConfigurationStore()
        config["password_storage"] = "environment"
        keys = InRamEncryptedKeyStore(config=config)
        self.assertTrue(keys.locked())
        keys.unlock("foobar")
        keys.password = "FOoo"
        with self.assertRaises(Exception):
            keys._decrypt_masterpassword()
        keys.lock()

        with self.assertRaises(WrongMasterPasswordException):
            keys.unlock("foobar2")

        with self.assertRaises(Exception):
            keys._get_encrypted_masterpassword()

        self.assertFalse(keys.unlocked())

        os.environ["UNLOCK"] = "foobar"
        self.assertTrue(keys.unlocked())

        self.assertFalse(keys.locked())



# -*- coding: utf-8 -*-
from nectar import Hive
from nectar.nodelist import NodeList


def get_hive_nodes():
    nodelist = NodeList()
    nodes = nodelist.get_hive_nodes()
    nodelist.update_nodes(blockchain_instance=Hive(node=nodes, num_retries=10))
    return nodelist.get_hive_nodes()
    # return "https://beta.openhive.network"


def get_steem_nodes():
    return "https://api.steemit.com"


def get_blurt_nodes():
    return "https://rpc.blurt.world"



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import time
from datetime import datetime, timedelta

from prettytable import PrettyTable

from nectar.account import Account
from nectar.block import Block
from nectar.blockchain import Blockchain
from nectar.nodelist import NodeList
from nectar.steem import Steem
from nectar.utils import formatTimedelta, parse_time
from nectarapi.exceptions import NumRetriesReached

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


if __name__ == "__main__":
    how_many_minutes = 10
    how_many_virtual_op = 10000
    max_batch_size = None
    threading = False
    thread_num = 16
    nodelist = NodeList()
    nodes = nodelist.get_nodes(normal=True, appbase=True, dev=True)
    t = PrettyTable(["node", "10 blockchain minutes", "10000 virtual account op", "version"])
    t.align = "l"
    for node in nodes:
        print("Current node:", node)
        try:
            stm = Steem(node=node, num_retries=3)
            blockchain = Blockchain(steem_instance=stm)
            account = Account("gtg", steem_instance=stm)
            virtual_op_count = account.virtual_op_count()
            blockchain_version = stm.get_blockchain_version()

            last_block_id = 19273700
            last_block = Block(last_block_id, steem_instance=stm)
            startTime = datetime.now()

            stopTime = last_block.time() + timedelta(seconds=how_many_minutes * 60)
            ltime = time.time()
            cnt = 0
            total_transaction = 0

            start_time = time.time()
            last_node = blockchain.steem.rpc.url

            for entry in blockchain.blocks(
                start=last_block_id,
                max_batch_size=max_batch_size,
                threading=threading,
                thread_num=thread_num,
            ):
                block_no = entry.identifier
                if "block" in entry:
                    trxs = entry["block"]["transactions"]
                else:
                    trxs = entry["transactions"]

                for tx in trxs:
                    for op in tx["operations"]:
                        total_transaction += 1
                if "block" in entry:
                    block_time = parse_time(entry["block"]["timestamp"])
                else:
                    block_time = parse_time(entry["timestamp"])

                if block_time > stopTime:
                    total_duration = formatTimedelta(datetime.now() - startTime)
                    last_block_id = block_no
                    avtran = total_transaction / (last_block_id - 19273700)
                    break
            start_time = time.time()

            stopOP = virtual_op_count - how_many_virtual_op + 1
            i = 0
            for acc_op in account.history_reverse(stop=stopOP):
                i += 1
            total_duration_acc = formatTimedelta(datetime.now() - startTime)

            print("* Processed %d blockchain minutes in %s" % (how_many_minutes, total_duration))
            print("* Processed %d account ops in %s" % (i, total_duration_acc))
            print("* blockchain version: %s" % (blockchain_version))
            t.add_row([node, total_duration, total_duration_acc, blockchain_version])
        except NumRetriesReached:
            print("NumRetriesReached")
            continue
        except Exception as e:
            print("Error: " + str(e))
            continue
    print(t)



# -*- coding: utf-8 -*-
import os
import unittest

from click.testing import CliRunner

from nectar.cli import cli
from nectar.instance import shared_steem_instance
from nectar.utils import import_pubkeys

from .nodes import get_hive_nodes

wif = "5Jt2wTfhUt5GkZHV1HYVfkEaJ6XnY8D2iA4qjtK9nnGXAhThM3w"
posting_key = "5Jh1Gtu2j4Yi16TfhoDmg8Qj3ULcgRi7A49JXdfUUTVPkaFaRKz"
memo_key = "5KPbCuocX26aMxN9CDPdUex4wCbfw9NoT5P7UhcqgDwxXa47bit"
pub_key = "STX52xMqKegLk4tdpNcUXU9Rw5DtdM9fxf3f12Gp55v1UjLX3ELZf"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_list = get_hive_nodes()

        # stm = shared_steem_instance()
        # stm.config.refreshBackup()
        runner = CliRunner()
        result = runner.invoke(cli, ["-o", "set", "default_vote_weight", "100"])
        if result.exit_code != 0:
            raise AssertionError(str(result))
        result = runner.invoke(cli, ["-o", "set", "default_account", "nectar"])
        if result.exit_code != 0:
            raise AssertionError(str(result))
        result = runner.invoke(cli, ["-o", "set", "nodes", str(cls.node_list)])
        if result.exit_code != 0:
            raise AssertionError(str(result))
        result = runner.invoke(cli, ["createwallet", "--wipe"], input="test\ntest\n")
        if result.exit_code != 0:
            raise AssertionError(str(result))
        result = runner.invoke(cli, ["addkey"], input="test\n" + wif + "\n")
        if result.exit_code != 0:
            raise AssertionError(str(result))
        result = runner.invoke(cli, ["addkey"], input="test\n" + posting_key + "\n")
        if result.exit_code != 0:
            raise AssertionError(str(result))
        result = runner.invoke(cli, ["addkey"], input="test\n" + memo_key + "\n")
        if result.exit_code != 0:
            raise AssertionError(str(result))

    @classmethod
    def tearDownClass(cls):
        stm = shared_steem_instance()
        stm.config.recover_with_latest_backup()
        runner = CliRunner()
        result = runner.invoke(cli, ["updatenodes", "--hive"])

    def test_balance(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["balance", "thecrazygm", "hive-nectar"])
        self.assertEqual(result.exit_code, 0)

    def test_interest(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "interest", "thecrazygm", "hive-nectar"])
        self.assertEqual(result.exit_code, 0)

    def test_config(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["config"])
        self.assertEqual(result.exit_code, 0)

    def test_addkey(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["createwallet", "--wipe"], input="test\ntest\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["addkey"], input="test\n" + wif + "\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["addkey"], input="test\n" + posting_key + "\n")
        self.assertEqual(result.exit_code, 0)

    def test_parsewif(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["parsewif"], input=wif + "\nexit\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["parsewif", "--unsafe-import-key", wif])
        self.assertEqual(result.exit_code, 0)

    def test_changerecovery(self):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["-dx", "changerecovery", "-a", "thecrazygm", "thecrazygm"], input=wif + "\nexit\n"
        )
        self.assertEqual(result.exit_code, 0)

    def test_delkey(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["delkey", "--confirm", pub_key], input="test\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["addkey"], input="test\n" + posting_key + "\n")
        # self.assertEqual(result.exit_code, 0)

    def test_listkeys(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["listkeys"])
        self.assertEqual(result.exit_code, 0)

    def test_listaccounts(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["listaccounts"])
        self.assertEqual(result.exit_code, 0)

    def test_info(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["info"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["info", "thecrazygm"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["info", "100"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["info", "--", "-1"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["info", pub_key])
        self.assertEqual(result.exit_code, 0)

    def test_info2(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["info", "--", "42725832:-1"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["info", "--", "42725832:1"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["info", "gtg"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["info", "@gtg/witness-gtg-log"])
        self.assertEqual(result.exit_code, 0)

    def test_changepassword(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["changewalletpassphrase"], input="test\ntest\ntest\n")
        self.assertEqual(result.exit_code, 0)

    def test_walletinfo(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["walletinfo"])
        self.assertEqual(result.exit_code, 0)

    def test_keygen(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["keygen"])
        self.assertEqual(result.exit_code, 0)

    def test_passwordgen(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["passwordgen"])
        self.assertEqual(result.exit_code, 0)
        data_dir = os.path.join(os.path.dirname(__file__), "data")
        file = os.path.join(data_dir, "drv-wif-idx100.txt")
        file2 = os.path.join(data_dir, "wif_pub_temp.json")
        result = runner.invoke(cli, ["passwordgen", "-a", "test", "-o", file, "-u", file2, "-w", 1])
        self.assertEqual(result.exit_code, 0)
        owner, active, posting, memo = import_pubkeys(file2)
        self.assertEqual(owner, "STM7d8DzUzjs5jbSkBVNctRaZFGe991MhzzTrqMoTVvZJ5oyZN7Cj")
        self.assertEqual(active, "STM7oADsCds97GqyEDY4cQC66brVrg7XHuRa2MLvYbuGrdKnNoQa6")
        self.assertEqual(posting, "STM5fpGcVwvUFF55EzWQ35oJeERcWvt4M9dXwehdpYmKaFCCqihL7")
        self.assertEqual(memo, "STM6A7DywWvMZRokxAK5CpTo8XAPKbrMennAs4ntwRFq5nj2jR7nG")
        os.remove(file2)

    def test_set(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-o", "set", "set_default_vote_weight", "100"])
        self.assertEqual(result.exit_code, 0)

    def test_upvote(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "upvote", "@steemit/firstpost"], input="test\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli, ["-dx", "upvote", "--weight", "100", "@steemit/firstpost"], input="test\n"
        )
        self.assertEqual(result.exit_code, 0)

    def test_downvote(self):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["-dx", "downvote", "--weight", "100", "@steemit/firstpost"], input="test\n"
        )
        self.assertEqual(result.exit_code, 0)

    def test_download(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "download", "-a", "steemit", "firstpost"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["-dx", "download", "@steemit/firstpost"])
        self.assertEqual(result.exit_code, 0)

    def test_transfer(self):
        stm = shared_steem_instance()
        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["-dx", "transfer", "thecrazygm", "1", stm.backed_token_symbol, "test"],
            input="test\n",
        )
        self.assertEqual(result.exit_code, 0)

    def test_powerdownroute(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "powerdownroute", "thecrazygm"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_convert(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "convert", "1"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_powerup(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "powerup", "1"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_powerdown(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "powerdown", "1e3"], input="test\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["-dx", "powerdown", "0"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_updatememokey(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "updatememokey"], input="test\ntest\ntest\n")
        self.assertEqual(result.exit_code, 0)

    def test_permissions(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["permissions", "thecrazygm"])
        self.assertEqual(result.exit_code, 0)

    def test_follower(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["follower", "thecrazygm"])
        self.assertEqual(result.exit_code, 0)

    def test_following(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["following", "thecrazygm"])
        self.assertEqual(result.exit_code, 0)

    def test_muter(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["muter", "thecrazygm"])
        self.assertEqual(result.exit_code, 0)

    def test_about(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["about"])
        self.assertEqual(result.exit_code, 0)

    def test_muting(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["muting", "nectar"])
        self.assertEqual(result.exit_code, 0)

    def test_allow_disallow(self):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["-dx", "allow", "--account", "thecrazygm", "--permission", "posting", "hive-nectar"],
            input="test\n",
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli,
            ["-dx", "disallow", "--account", "thecrazygm", "--permission", "posting", "rewarding"],
            input="test\n",
        )
        self.assertEqual(result.exit_code, 0)

    def test_witnesses(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["witnesses"])
        self.assertEqual(result.exit_code, 0)

    @unittest.skip
    def test_votes(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["votes", "--direction", "out", "fullnodeupdate"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["votes", "--direction", "in", "fullnodeupdate"])
        self.assertEqual(result.exit_code, 0)

    def test_approvewitness(self):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["-dx", "approvewitness", "-a", "hive-nectar", "thecrazygm"], input="test\n"
        )
        self.assertEqual(result.exit_code, 0)

    def test_disapprovewitness(self):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["-dx", "disapprovewitness", "-a", "hive-nectar", "thecrazygm"], input="test\n"
        )
        self.assertEqual(result.exit_code, 0)

    def test_addproxy(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "setproxy", "-a", "hive-nectar", "thecrazygm"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_delproxy(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "delproxy", "-a", "fullnodeupdate"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_newaccount(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "newaccount", "nectar3"], input="test\ntest\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli,
            [
                "-dx",
                "newaccount",
                "--owner",
                "STM7mLs2hns87f7kbf3o2HBqNoEaXiTeeU89eVF6iUCrMQJFzBsPo",
                "--active",
                "STM7rUmnpnCp9oZqMQeRKDB7GvXTM9KFvhzbA3AKcabgTBfQZgHZp",
                "--posting",
                "STM6qGWHsCpmHbphnQbS2yfhvhJXDUVDwnsbnrMZkTqfnkNEZRoLP",
                "--memo",
                "STM8Wvi74GYzBKgnUmiLvptzvxmPtXfjGPJL8QY3rebecXaxGGQyV",
                "nectar3",
            ],
            input="test\ntest\n",
        )
        self.assertEqual(result.exit_code, 0)

    def test_changekeys(self):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "-dx",
                "changekeys",
                "--owner",
                "STM7mLs2hns87f7kbf3o2HBqNoEaXiTeeU89eVF6iUCrMQJFzBsPo",
                "--active",
                "STM7rUmnpnCp9oZqMQeRKDB7GvXTM9KFvhzbA3AKcabgTBfQZgHZp",
                "--posting",
                "STM6qGWHsCpmHbphnQbS2yfhvhJXDUVDwnsbnrMZkTqfnkNEZRoLP",
                "--memo",
                "STM8Wvi74GYzBKgnUmiLvptzvxmPtXfjGPJL8QY3rebecXaxGGQyV",
                "nectar",
            ],
            input="test\ntest\n",
        )
        self.assertEqual(result.exit_code, 0)

    @unittest.skip
    def test_importaccount(self):
        runner = CliRunner()
        runner.invoke(cli, ["-o", "set", "nodes", str(self.node_list)])
        result = runner.invoke(
            cli,
            ["importaccount", "--roles", '["owner", "active", "posting", "memo"]', "nectar2"],
            input="test\numybjvCafrt8LdoCjEimQiQ4\n",
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli,
            ["delkey", "--confirm", "STX7mLs2hns87f7kbf3o2HBqNoEaXiTeeU89eVF6iUCrMQJFzBsPo"],
            input="test\n",
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli,
            ["delkey", "--confirm", "STX7rUmnpnCp9oZqMQeRKDB7GvXTM9KFvhzbA3AKcabgTBfQZgHZp"],
            input="test\n",
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli,
            ["delkey", "--confirm", "STX6qGWHsCpmHbphnQbS2yfhvhJXDUVDwnsbnrMZkTqfnkNEZRoLP"],
            input="test\n",
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli,
            ["delkey", "--confirm", "STX8Wvi74GYzBKgnUmiLvptzvxmPtXfjGPJL8QY3rebecXaxGGQyV"],
            input="test\n",
        )
        self.assertEqual(result.exit_code, 0)

    def test_orderbook(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["orderbook"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["orderbook", "--show-date"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["orderbook", "--chart"])
        self.assertEqual(result.exit_code, 0)

    def test_buy(self):
        stm = shared_steem_instance()
        runner = CliRunner()
        result = runner.invoke(
            cli, ["-dt", "-x", "buy", "1", stm.token_symbol, "2.2"], input="test\n"
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["-dt", "-x", "buy", "1", stm.token_symbol], input="y\ntest\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli, ["-dt", "-x", "buy", "1", stm.backed_token_symbol, "2.2"], input="test\n"
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli, ["-dt", "-x", "buy", "1", stm.backed_token_symbol], input="y\ntest\n"
        )
        self.assertEqual(result.exit_code, 0)

    def test_sell(self):
        stm = shared_steem_instance()
        runner = CliRunner()
        result = runner.invoke(
            cli, ["-dt", "-x", "sell", "1", stm.token_symbol, "2.2"], input="test\n"
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli, ["-dt", "-x", "sell", "1", stm.backed_token_symbol, "2.2"], input="test\n"
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["-dt", "-x", "sell", "1", stm.token_symbol], input="y\ntest\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli, ["-dt", "-x", "sell", "1", stm.backed_token_symbol], input="y\ntest\n"
        )
        self.assertEqual(result.exit_code, 0)

    def test_cancel(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "cancel", "5"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_openorders(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["openorders"])
        self.assertEqual(result.exit_code, 0)

    def test_reblog(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dto", "reblog", "@steemit/firstpost"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_follow_unfollow(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dto", "follow", "hive-nectar"], input="test\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["-dto", "unfollow", "hive-nectar"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_mute_unmute(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dto", "mute", "hive-nectar"], input="test\n")
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["-dto", "unfollow", "hive-nectar"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_witnesscreate(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "witnesscreate", "nectar", pub_key], input="test\n")

    def test_witnessupdate(self):
        runner = CliRunner()
        runner.invoke(
            cli,
            [
                "-dx",
                "witnessupdate",
                "gtg",
                "--maximum_block_size",
                65000,
                "--account_creation_fee",
                0.1,
                "--sbd_interest_rate",
                0,
                "--url",
                "https://google.de",
                "--signing_key",
                wif,
            ],
        )

    def test_profile(self):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["-dx", "setprofile", "url", "https://google.de"], input="test\n"
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["-dx", "delprofile", "url"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_claimreward(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "claimreward"], input="test\n")
        result = runner.invoke(cli, ["-dx", "claimreward", "--claim_all_steem"], input="test\n")
        result = runner.invoke(cli, ["-dx", "claimreward", "--claim_all_sbd"], input="test\n")
        result = runner.invoke(cli, ["-dx", "claimreward", "--claim_all_vests"], input="test\n")
        self.assertEqual(result.exit_code, 0)

    def test_claimaccount(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "claimaccount", "thecrazygm"])
        result = runner.invoke(cli, ["-dx", "claimaccount", "-n", "2", "thecrazygm"])
        self.assertEqual(result.exit_code, 0)

    def test_power(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["power", "thecrazygm"])
        self.assertEqual(result.exit_code, 0)

    def test_history(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["history", "thecrazygm"])
        self.assertEqual(result.exit_code, 0)

    def test_draw(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["draw"])
        self.assertEqual(result.exit_code, 0)

    def test_witnessenable(self):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["-dx", "witnessenable", "thecrazygm", "STM1111111111111111111111111111111114T1Anm"]
        )
        self.assertEqual(result.exit_code, 0)

    def test_witnessdisable(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-dx", "witnessdisable", "thecrazygm"])
        self.assertEqual(result.exit_code, 0)

    def test_nextnode(self):
        runner = CliRunner()
        runner.invoke(cli, ["-o", "set", "nodes", self.node_list])
        result = runner.invoke(cli, ["-o", "nextnode"])
        self.assertEqual(result.exit_code, 0)
        runner.invoke(cli, ["-o", "set", "nodes", str(self.node_list)])

    def test_pingnode(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["pingnode"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["pingnode", "--sort"])
        self.assertEqual(result.exit_code, 0)

    def test_updatenodes(self):
        runner = CliRunner()
        runner.invoke(cli, ["-o", "set", "nodes", self.node_list])
        result = runner.invoke(cli, ["updatenodes", "--hive", "--test"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["updatenodes", "--steem"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["updatenodes"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["updatenodes", "--hive"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["updatenodes"])
        self.assertEqual(result.exit_code, 0)
        runner.invoke(cli, ["-o", "set", "nodes", str(self.node_list)])

    def test_currentnode(self):
        runner = CliRunner()
        runner.invoke(cli, ["-o", "set", "nodes", self.node_list])
        result = runner.invoke(cli, ["currentnode"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["currentnode", "--url"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["currentnode", "--version"])
        self.assertEqual(result.exit_code, 0)
        runner.invoke(cli, ["-o", "set", "nodes", str(self.node_list)])

    def test_ticker(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["ticker"])
        self.assertEqual(result.exit_code, 0)

    def test_pricehistory(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["pricehistory"])
        self.assertEqual(result.exit_code, 0)

    def test_notifications(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["notifications", "fullnodeupdate"])
        self.assertEqual(result.exit_code, 0)

    def test_pending(self):
        runner = CliRunner()
        account_name = "fullnodeupdate"
        result = runner.invoke(cli, ["pending", account_name])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["pending", "--post", "--comment", account_name])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli, ["pending", "--curation", "--permlink", "--days", "1", account_name]
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli,
            [
                "pending",
                "--post",
                "--comment",
                "--author",
                "--permlink",
                "--length",
                "30",
                "--days",
                "1",
                account_name,
            ],
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli, ["pending", "--post", "--author", "--title", "--days", "1", account_name]
        )
        self.assertEqual(result.exit_code, 0)

    def test_rewards(self):
        runner = CliRunner()
        account_name = "fullnodeupdate"
        result = runner.invoke(cli, ["rewards", account_name])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["rewards", "--post", "--comment", "--curation", account_name])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli, ["rewards", "--post", "--comment", "--curation", "--permlink", account_name]
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli, ["rewards", "--post", "--comment", "--curation", "--author", account_name]
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli, ["rewards", "--post", "--comment", "--author", "--title", account_name]
        )
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(
            cli,
            [
                "rewards",
                "--post",
                "--comment",
                "--author",
                "--permlink",
                "--length",
                "30",
                account_name,
            ],
        )
        self.assertEqual(result.exit_code, 0)

    def test_curation(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["curation", "@gtg/witness-gtg-log"])
        self.assertEqual(result.exit_code, 0)

    def test_verify(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["verify", "--trx", "3", "25304468"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["verify", "--trx", "5", "25304468"])
        self.assertEqual(result.exit_code, 0)
        result = runner.invoke(cli, ["verify", "--trx", "0"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.exit_code, 0)

    def test_tradehistory(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["tradehistory"])
        self.assertEqual(result.exit_code, 0)



#!/usr/bin/python
import sys
from time import sleep

from nectar import Steem
from nectar.witness import Witness, WitnessesRankedByVote


def convert_block_diff_to_time_string(block_diff_est):
    next_block_s = int(block_diff_est) * 3
    next_block_min = next_block_s / 60
    next_block_h = next_block_min / 60
    next_block_d = next_block_h / 24
    time_diff_est = ""
    if next_block_d > 1:
        time_diff_est = "%.2f days" % next_block_d
    elif next_block_h > 1:
        time_diff_est = "%.2f hours" % next_block_h
    elif next_block_min > 1:
        time_diff_est = "%.2f minutes" % next_block_min
    else:
        time_diff_est = "%.2f seconds" % next_block_s
    return time_diff_est


if __name__ == "__main__":
    if len(sys.argv) != 2:
        witness = "thecrazygm"
    else:
        witness = sys.argv[1]
    stm = Steem()
    witness = Witness(witness, steem_instance=stm)

    witness_schedule = stm.get_witness_schedule()
    config = stm.get_config()
    if "VIRTUAL_SCHEDULE_LAP_LENGTH2" in config:
        lap_length = int(config["VIRTUAL_SCHEDULE_LAP_LENGTH2"])
    else:
        lap_length = int(config["STEEM_VIRTUAL_SCHEDULE_LAP_LENGTH2"])
    witnesses = WitnessesRankedByVote(limit=250, steem_instance=stm)
    vote_sum = witnesses.get_votes_sum()

    virtual_time_to_block_num = int(witness_schedule["num_scheduled_witnesses"]) / (
        lap_length / (vote_sum + 1)
    )
    while True:
        witness.refresh()
        witness_schedule = stm.get_witness_schedule(use_stored_data=False)

        witness_json = witness.json()
        virtual_diff = int(witness_json["virtual_scheduled_time"]) - int(
            witness_schedule["current_virtual_time"]
        )
        block_diff_est = virtual_diff * virtual_time_to_block_num

        time_diff_est = convert_block_diff_to_time_string(block_diff_est)

        sys.stdout.write("\r Next block for %s in %s" % (witness["owner"], time_diff_est))
        sleep(30)



# -*- coding: utf-8 -*-
import json
from datetime import date, datetime, timezone

from prettytable import PrettyTable

from nectarapi.exceptions import InvalidParameters, UnknownKey

from .account import Account
from .blockchainobject import BlockchainObject
from .comment import Comment
from .exceptions import VoteDoesNotExistsException
from .instance import shared_blockchain_instance
from .utils import (
    addTzInfo,
    construct_authorperm,
    construct_authorpermvoter,
    formatTimeString,
    reputation_to_score,
    resolve_authorperm,
    resolve_authorpermvoter,
)


class Vote(BlockchainObject):
    """Read data about a Vote in the chain

    :param str authorperm: perm link to post/comment
    :param nectar.nectar.nectar blockchain_instance: nectar
        instance to use when accesing a RPC
    """

    def __init__(
        self, voter, authorperm=None, lazy=False, full=False, blockchain_instance=None, **kwargs
    ):
        self.full = full
        self.lazy = lazy
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if isinstance(voter, str) and authorperm is not None:
            [author, permlink] = resolve_authorperm(authorperm)
            self["voter"] = voter
            self["author"] = author
            self["permlink"] = permlink
            authorpermvoter = construct_authorpermvoter(author, permlink, voter)
            self["authorpermvoter"] = authorpermvoter
        elif (
            isinstance(voter, dict)
            and "author" in voter
            and "permlink" in voter
            and "voter" in voter
        ):
            authorpermvoter = voter
            authorpermvoter["authorpermvoter"] = construct_authorpermvoter(
                voter["author"], voter["permlink"], voter["voter"]
            )
            authorpermvoter = self._parse_json_data(authorpermvoter)
        elif isinstance(voter, dict) and "authorperm" in voter and authorperm is not None:
            [author, permlink] = resolve_authorperm(voter["authorperm"])
            authorpermvoter = voter
            authorpermvoter["voter"] = authorperm
            authorpermvoter["author"] = author
            authorpermvoter["permlink"] = permlink
            authorpermvoter["authorpermvoter"] = construct_authorpermvoter(
                author, permlink, authorperm
            )
            authorpermvoter = self._parse_json_data(authorpermvoter)
        elif isinstance(voter, dict) and "voter" in voter and authorperm is not None:
            [author, permlink] = resolve_authorperm(authorperm)
            authorpermvoter = voter
            authorpermvoter["author"] = author
            authorpermvoter["permlink"] = permlink
            authorpermvoter["authorpermvoter"] = construct_authorpermvoter(
                author, permlink, voter["voter"]
            )
            authorpermvoter = self._parse_json_data(authorpermvoter)
        else:
            authorpermvoter = voter
            [author, permlink, voter] = resolve_authorpermvoter(authorpermvoter)
            self["author"] = author
            self["permlink"] = permlink

        super(Vote, self).__init__(
            authorpermvoter,
            id_item="authorpermvoter",
            lazy=lazy,
            full=full,
            blockchain_instance=blockchain_instance,
        )

    def refresh(self):
        if self.identifier is None:
            return
        if not self.blockchain.is_connected():
            return
        [author, permlink, voter] = resolve_authorpermvoter(self.identifier)
        try:
            self.blockchain.rpc.set_next_node_on_empty_reply(True)
            if self.blockchain.rpc.get_use_appbase():
                try:
                    votes = self.blockchain.rpc.get_active_votes(
                        {"author": author, "permlink": permlink}, api="condenser"
                    )["votes"]
                except InvalidParameters:
                    raise VoteDoesNotExistsException(self.identifier)
                except Exception:  # Fallback to condenser API
                    votes = self.blockchain.rpc.get_active_votes(author, permlink, api="condenser")
            else:
                votes = self.blockchain.rpc.get_active_votes(author, permlink, api="condenser")
        except UnknownKey:
            raise VoteDoesNotExistsException(self.identifier)

        vote = None
        if votes is not None:
            for x in votes:
                if x["voter"] == voter:
                    vote = x
        if not vote:
            raise VoteDoesNotExistsException(self.identifier)
        vote = self._parse_json_data(vote)
        vote["authorpermvoter"] = construct_authorpermvoter(author, permlink, voter)
        super(Vote, self).__init__(
            vote,
            id_item="authorpermvoter",
            lazy=self.lazy,
            full=self.full,
            blockchain_instance=self.blockchain,
        )

    def _parse_json_data(self, vote):
        parse_int = [
            "rshares",
            "reputation",
        ]
        for p in parse_int:
            if p in vote and isinstance(vote.get(p), str):
                vote[p] = int(vote.get(p, "0"))

        if "time" in vote and isinstance(vote.get("time"), str) and vote.get("time") != "":
            vote["time"] = formatTimeString(vote.get("time", "1970-01-01T00:00:00"))
        elif (
            "timestamp" in vote
            and isinstance(vote.get("timestamp"), str)
            and vote.get("timestamp") != ""
        ):
            vote["time"] = formatTimeString(vote.get("timestamp", "1970-01-01T00:00:00"))
        elif (
            "last_update" in vote
            and isinstance(vote.get("last_update"), str)
            and vote.get("last_update") != ""
        ):
            vote["last_update"] = formatTimeString(vote.get("last_update", "1970-01-01T00:00:00"))
        else:
            vote["time"] = formatTimeString("1970-01-01T00:00:00")
        return vote

    def json(self):
        output = self.copy()
        if "author" in output:
            output.pop("author")
        if "permlink" in output:
            output.pop("permlink")
        parse_times = ["time"]
        for p in parse_times:
            if p in output:
                p_date = output.get(p, datetime(1970, 1, 1, 0, 0))
                if isinstance(p_date, (datetime, date)):
                    output[p] = formatTimeString(p_date)
                else:
                    output[p] = p_date
        parse_int = [
            "rshares",
            "reputation",
        ]
        for p in parse_int:
            if p in output and isinstance(output[p], int):
                output[p] = str(output[p])
        return json.loads(str(json.dumps(output)))

    @property
    def voter(self):
        return self["voter"]

    @property
    def authorperm(self):
        if "authorperm" in self:
            return self["authorperm"]
        elif "authorpermvoter" in self:
            [author, permlink, voter] = resolve_authorpermvoter(self["authorpermvoter"])
            return construct_authorperm(author, permlink)
        elif "author" in self and "permlink" in self:
            return construct_authorperm(self["author"], self["permlink"])
        else:
            return ""

    @property
    def votee(self):
        votee = ""
        authorperm = self.get("authorperm", "")
        authorpermvoter = self.get("authorpermvoter", "")
        if authorperm != "":
            votee = resolve_authorperm(authorperm)[0]
        elif authorpermvoter != "":
            votee = resolve_authorpermvoter(authorpermvoter)[0]
        return votee

    @property
    def weight(self):
        return self["weight"]

    @property
    def sbd(self):
        return self.blockchain.rshares_to_sbd(int(self.get("rshares", 0)))

    @property
    def hbd(self):
        return self.blockchain.rshares_to_hbd(int(self.get("rshares", 0)))

    @property
    def token_backed_dollar(self):
        from nectar import Hive

        if isinstance(self.blockchain, Hive):
            return self.blockchain.rshares_to_hbd(int(self.get("rshares", 0)))
        else:
            return self.blockchain.rshares_to_sbd(int(self.get("rshares", 0)))

    @property
    def rshares(self):
        return int(self.get("rshares", 0))

    @property
    def percent(self):
        return self.get("percent", 0)

    @property
    def reputation(self):
        return self.get("reputation", 0)

    @property
    def rep(self):
        return reputation_to_score(int(self.reputation))

    @property
    def time(self):
        return self["time"]


class VotesObject(list):
    def get_sorted_list(self, sort_key="time", reverse=True):
        sortedList = sorted(
            self,
            key=lambda x: (datetime.now(timezone.utc) - x.time).total_seconds(),
            reverse=reverse,
        )
        return sortedList

    def printAsTable(
        self,
        voter=None,
        votee=None,
        start=None,
        stop=None,
        start_percent=None,
        stop_percent=None,
        sort_key="time",
        reverse=True,
        allow_refresh=True,
        return_str=False,
        **kwargs,
    ):
        table_header = ["Voter", "Votee", "SBD/HBD", "Time", "Rshares", "Percent", "Weight"]
        t = PrettyTable(table_header)
        t.align = "l"
        start = addTzInfo(start)
        stop = addTzInfo(stop)
        for vote in self.get_sorted_list(sort_key=sort_key, reverse=reverse):
            if not allow_refresh:
                vote.cached = True

            d_time = vote.time
            if d_time != formatTimeString("1970-01-01T00:00:00"):
                td = datetime.now(timezone.utc) - d_time
                timestr = (
                    str(td.days)
                    + " days "
                    + str(td.seconds // 3600)
                    + ":"
                    + str((td.seconds // 60) % 60)
                )
            else:
                start = None
                stop = None
                timestr = ""

            percent = vote.get("percent", "")
            if percent == "":
                start_percent = None
                stop_percent = None
            if (
                (start is None or d_time >= start)
                and (stop is None or d_time <= stop)
                and (start_percent is None or percent >= start_percent)
                and (stop_percent is None or percent <= stop_percent)
                and (voter is None or vote["voter"] == voter)
                and (votee is None or vote.votee == votee)
            ):
                percent = vote.get("percent", "")
                if percent == "":
                    percent = vote.get("vote_percent", "")
                t.add_row(
                    [
                        vote["voter"],
                        vote.votee,
                        str(round(vote.token_backed_dollar, 2)).ljust(5) + "$",
                        timestr,
                        vote.get("rshares", ""),
                        str(percent),
                        str(vote["weight"]),
                    ]
                )

        if return_str:
            return t.get_string(**kwargs)
        else:
            print(t.get_string(**kwargs))

    def get_list(
        self,
        var="voter",
        voter=None,
        votee=None,
        start=None,
        stop=None,
        start_percent=None,
        stop_percent=None,
        sort_key="time",
        reverse=True,
    ):
        vote_list = []
        start = addTzInfo(start)
        stop = addTzInfo(stop)
        for vote in self.get_sorted_list(sort_key=sort_key, reverse=reverse):
            d_time = vote.time
            if d_time != formatTimeString("1970-01-01T00:00:00"):
                start = None
                stop = None
            percent = vote.get("percent", "")
            if percent == "":
                percent = vote.get("vote_percent", "")
            if percent == "":
                start_percent = None
                stop_percent = None
            if (
                (start is None or d_time >= start)
                and (stop is None or d_time <= stop)
                and (start_percent is None or percent >= start_percent)
                and (stop_percent is None or percent <= stop_percent)
                and (voter is None or vote["voter"] == voter)
                and (votee is None or vote.votee == votee)
            ):
                v = ""
                if var == "voter":
                    v = vote["voter"]
                elif var == "votee":
                    v = vote.votee
                elif var == "sbd" or var == "hbd":
                    v = vote.token_backed_dollar
                elif var == "time":
                    v = d_time
                elif var == "rshares":
                    v = vote.get("rshares", 0)
                elif var == "percent":
                    v = percent
                elif var == "weight":
                    v = vote["weight"]
                vote_list.append(v)
        return vote_list

    def print_stats(self, return_str=False, **kwargs):
        # Using built-in timezone support
        table_header = ["voter", "votee", "sbd/hbd", "time", "rshares", "percent", "weight"]
        t = PrettyTable(table_header)
        t.align = "l"

    def __contains__(self, item):
        if isinstance(item, Account):
            name = item["name"]
            authorperm = ""
        elif isinstance(item, Comment):
            authorperm = item.authorperm
            name = ""
        else:
            name = item
            authorperm = item

        return (
            any([name == x.voter for x in self])
            or any([name == x.votee for x in self])
            or any([authorperm == x.authorperm for x in self])
        )

    def __str__(self):
        return self.printAsTable(return_str=True)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, str(self.identifier))


class ActiveVotes(VotesObject):
    """Obtain a list of votes for a post

    :param str authorperm: authorperm link
    :param Steem steem_instance: Steem() instance to use when accesing a RPC
    """

    def __init__(self, authorperm, lazy=False, full=False, blockchain_instance=None, **kwargs):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        votes = None
        if not self.blockchain.is_connected():
            return None
        self.blockchain.rpc.set_next_node_on_empty_reply(False)

        if isinstance(authorperm, Comment):
            # if 'active_votes' in authorperm and len(authorperm["active_votes"]) > 0:
            #    votes = authorperm["active_votes"]
            if self.blockchain.rpc.get_use_appbase():
                self.blockchain.rpc.set_next_node_on_empty_reply(False)
                from nectarapi.exceptions import InvalidParameters

                try:
                    votes = self.blockchain.rpc.get_active_votes(
                        authorperm["author"], authorperm["permlink"], api="condenser"
                    )
                except InvalidParameters:
                    raise VoteDoesNotExistsException(
                        construct_authorperm(authorperm["author"], authorperm["permlink"])
                    )
                except Exception:  # Fallback to tags API
                    votes = self.blockchain.rpc.get_active_votes(
                        {"author": authorperm["author"], "permlink": authorperm["permlink"]},
                        api="tags",
                    )["votes"]
            else:
                votes = self.blockchain.rpc.get_active_votes(
                    authorperm["author"], authorperm["permlink"], api="condenser"
                )
            authorperm = authorperm["authorperm"]
        elif isinstance(authorperm, str):
            [author, permlink] = resolve_authorperm(authorperm)
            if self.blockchain.rpc.get_use_appbase():
                self.blockchain.rpc.set_next_node_on_empty_reply(False)
                from nectarapi.exceptions import InvalidParameters

                try:
                    votes = self.blockchain.rpc.get_active_votes(author, permlink, api="condenser")
                except InvalidParameters:
                    raise VoteDoesNotExistsException(construct_authorperm(author, permlink))
                except Exception:  # Fallback to tags API
                    votes = self.blockchain.rpc.get_active_votes(
                        {"author": author, "permlink": permlink}, api="tags"
                    )["votes"]
            else:
                votes = self.blockchain.rpc.get_active_votes(author, permlink, api="condenser")
        elif isinstance(authorperm, list):
            votes = authorperm
            authorperm = None
        elif isinstance(authorperm, dict):
            votes = authorperm["active_votes"]
            authorperm = authorperm["authorperm"]
        if votes is None:
            return
        self.identifier = authorperm
        super(ActiveVotes, self).__init__(
            [
                Vote(
                    x,
                    authorperm=authorperm,
                    lazy=lazy,
                    full=full,
                    blockchain_instance=self.blockchain,
                )
                for x in votes
            ]
        )


class AccountVotes(VotesObject):
    """Obtain a list of votes for an account
    Lists the last 100+ votes on the given account.

    :param str account: Account name
    :param Steem steem_instance: Steem() instance to use when accesing a RPC
    """

    def __init__(
        self,
        account,
        start=None,
        stop=None,
        raw_data=False,
        lazy=False,
        full=False,
        blockchain_instance=None,
        **kwargs,
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        start = addTzInfo(start)
        stop = addTzInfo(stop)
        account = Account(account, blockchain_instance=self.blockchain)
        votes = account.get_account_votes()
        self.identifier = account["name"]
        vote_list = []
        if votes is None:
            votes = []
        for x in votes:
            time = x.get("time", "")
            if time == "":
                time = x.get("last_update", "")
                if time != "":
                    x["time"] = time
            if time != "" and isinstance(time, str):
                d_time = formatTimeString(time)
            elif isinstance(time, datetime):
                d_time = time
            else:
                d_time = addTzInfo(datetime(1970, 1, 1, 0, 0, 0))
            if (start is None or d_time >= start) and (stop is None or d_time <= stop):
                if not raw_data:
                    vote_list.append(
                        Vote(
                            x,
                            authorperm=account["name"],
                            lazy=lazy,
                            full=full,
                            blockchain_instance=self.blockchain,
                        )
                    )
                else:
                    vote_list.append(x)

        super(AccountVotes, self).__init__(vote_list)



# -*- coding: utf-8 -*-
import unittest

from nectar import Steem, constants

from .nodes import get_steem_nodes

wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.appbase = Steem(
            node=get_steem_nodes(),
            nobroadcast=True,
            bundle=False,
            # Overwrite wallet to use this list of wifs only
            keys={"active": wif},
            num_retries=10,
        )

    def test_constants(self):
        stm = self.appbase
        steem_conf = stm.get_config()
        if "STEEM_100_PERCENT" in steem_conf:
            STEEM_100_PERCENT = steem_conf["STEEM_100_PERCENT"]
        else:
            STEEM_100_PERCENT = steem_conf["STEEMIT_100_PERCENT"]
        self.assertEqual(constants.STEEM_100_PERCENT, STEEM_100_PERCENT)

        if "STEEM_1_PERCENT" in steem_conf:
            STEEM_1_PERCENT = steem_conf["STEEM_1_PERCENT"]
        else:
            STEEM_1_PERCENT = steem_conf["STEEMIT_1_PERCENT"]
        self.assertEqual(constants.STEEM_1_PERCENT, STEEM_1_PERCENT)

        if "STEEM_REVERSE_AUCTION_WINDOW_SECONDS" in steem_conf:
            STEEM_REVERSE_AUCTION_WINDOW_SECONDS = steem_conf[
                "STEEM_REVERSE_AUCTION_WINDOW_SECONDS"
            ]
        elif "STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF6" in steem_conf:
            STEEM_REVERSE_AUCTION_WINDOW_SECONDS = steem_conf[
                "STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF6"
            ]
        else:
            STEEM_REVERSE_AUCTION_WINDOW_SECONDS = steem_conf[
                "STEEMIT_REVERSE_AUCTION_WINDOW_SECONDS"
            ]
        self.assertEqual(
            constants.STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF6, STEEM_REVERSE_AUCTION_WINDOW_SECONDS
        )

        if "STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF20" in steem_conf:
            self.assertEqual(
                constants.STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF20,
                steem_conf["STEEM_REVERSE_AUCTION_WINDOW_SECONDS_HF20"],
            )

        if "STEEM_VOTE_DUST_THRESHOLD" in steem_conf:
            self.assertEqual(
                constants.STEEM_VOTE_DUST_THRESHOLD, steem_conf["STEEM_VOTE_DUST_THRESHOLD"]
            )

        if "STEEM_VOTE_REGENERATION_SECONDS" in steem_conf:
            STEEM_VOTE_REGENERATION_SECONDS = steem_conf["STEEM_VOTE_REGENERATION_SECONDS"]
            self.assertEqual(
                constants.STEEM_VOTE_REGENERATION_SECONDS, STEEM_VOTE_REGENERATION_SECONDS
            )
        elif "STEEM_VOTING_MANA_REGENERATION_SECONDS" in steem_conf:
            STEEM_VOTING_MANA_REGENERATION_SECONDS = steem_conf[
                "STEEM_VOTING_MANA_REGENERATION_SECONDS"
            ]
            self.assertEqual(
                constants.STEEM_VOTING_MANA_REGENERATION_SECONDS,
                STEEM_VOTING_MANA_REGENERATION_SECONDS,
            )
        else:
            STEEM_VOTE_REGENERATION_SECONDS = steem_conf["STEEMIT_VOTE_REGENERATION_SECONDS"]
            self.assertEqual(
                constants.STEEM_VOTE_REGENERATION_SECONDS, STEEM_VOTE_REGENERATION_SECONDS
            )

        if "STEEM_ROOT_POST_PARENT" in steem_conf:
            STEEM_ROOT_POST_PARENT = steem_conf["STEEM_ROOT_POST_PARENT"]
        else:
            STEEM_ROOT_POST_PARENT = steem_conf["STEEMIT_ROOT_POST_PARENT"]
        self.assertEqual(constants.STEEM_ROOT_POST_PARENT, STEEM_ROOT_POST_PARENT)



# This Python file uses the following encoding: utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

TEST_UNICODE_STR = "ℝεα∂@ßʟ℮ ☂ℯṧт υηḯ¢☺ḓ℮"
# Tk icon as a .gif:
TEST_BYTE_STR = b"GIF89a\x0e\x00\x0b\x00\x80\xff\x00\xff\x00\x00\xc0\xc0\xc0!\xf9\x04\x01\x00\x00\x01\x00,\x00\x00\x00\x00\x0e\x00\x0b\x00@\x02\x1f\x0c\x8e\x10\xbb\xcan\x90\x99\xaf&\xd8\x1a\xce\x9ar\x06F\xd7\xf1\x90\xa1c\x9e\xe8\x84\x99\x89\x97\xa2J\x01\x00;\x1a\x14\x00;;\xba\nD\x14\x00\x00;;"


class Testcases(unittest.TestCase):
    def test_bytes_encoding_arg(self):
        """
        The bytes class has changed in Python 3 to accept an
        additional argument in the constructor: encoding.
        It would be nice to support this without breaking the
        isinstance(..., bytes) test below.
        """
        u = "Unicode string: \u5b54\u5b50"
        b = bytes(u, encoding="utf-8")
        self.assertEqual(b, u.encode("utf-8"))

    def test_bytes_encoding_arg_non_kwarg(self):
        """
        As above, but with a positional argument
        """
        u = "Unicode string: \u5b54\u5b50"
        b = bytes(u, "utf-8")
        self.assertEqual(b, u.encode("utf-8"))

    def test_bytes_int(self):
        """
        In Py3, bytes(int) -> bytes object of size given by the parameter initialized with null
        """
        self.assertEqual(bytes(5), b"\x00\x00\x00\x00\x00")
        # Test using newint:
        self.assertEqual(bytes(int(5)), b"\x00\x00\x00\x00\x00")
        self.assertTrue(isinstance(bytes(int(5)), bytes))

    def test_bytes_iterable_of_ints(self):
        self.assertEqual(bytes([65, 66, 67]), b"ABC")
        self.assertEqual(bytes([int(120), int(121), int(122)]), b"xyz")

    def test_bytes_bytes(self):
        self.assertEqual(bytes(b"ABC"), b"ABC")

    def test_bytes_is_bytes(self):
        b = bytes(b"ABC")
        self.assertTrue(bytes(b) is b)
        self.assertEqual(repr(bytes(b)), "b'ABC'")

    def test_empty_bytes(self):
        b = bytes()
        self.assertEqual(b, b"")

    def test_isinstance_bytes(self):
        self.assertIsInstance(bytes(b"blah"), bytes)

    def test_isinstance_bytes_subclass(self):
        """
        Issue #89
        """
        value = bytes(b"abc")

        class Magic:
            def __bytes__(self):
                return bytes(b"abc")

        self.assertEqual(value, bytes(Magic()))

    def test_isinstance_oldbytestrings_bytes(self):
        """
        Watch out for this. Byte-strings produced in various places in Py2
        are of type 'str'. With 'from future.builtins import bytes', 'bytes'
        is redefined to be a subclass of 'str', not just an alias for 'str'.
        """
        self.assertIsInstance(b"blah", bytes)  # not with the redefined bytes obj
        self.assertIsInstance("blah".encode("utf-8"), bytes)  # not with the redefined bytes obj

    def test_bytes_getitem(self):
        b = bytes(b"ABCD")
        self.assertEqual(b[0], 65)
        self.assertEqual(b[-1], 68)
        self.assertEqual(b[0:1], b"A")
        self.assertEqual(b[:], b"ABCD")

    def test_int(self):
        a = int(5)
        b = int(10)
        self.assertIsInstance(a, int)
        self.assertIsInstance(b, int)

    def test_str(self):
        a = "abc"
        self.assertIsInstance(a, str)

    def test_chr(self):
        BASE58_ALPHABET = b"123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
        self.assertEqual(BASE58_ALPHABET.find(chr(4)), -1)
        self.assertEqual(BASE58_ALPHABET.find(b"Z"), 32)
        self.assertEqual(BASE58_ALPHABET.find(bytes("Z", "ascii")), 32)


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range
from timeit import default_timer as timer

from steembase import operations as steemOperations
from steembase.account import PrivateKey as steemPrivateKey
from steembase.transactions import SignedTransaction as steemSignedTransaction

from nectar.amount import Amount
from nectar.steem import Steem
from nectarbase import operations
from nectarbase.objects import Operation
from nectarbase.signedtransactions import Signed_Transaction
from nectargraphenebase.account import PrivateKey


class nectarTest(object):
    def setup(self):
        self.prefix = "STEEM"
        self.default_prefix = "STM"
        self.wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"
        self.ref_block_num = 34294
        self.ref_block_prefix = 3707022213
        self.expiration = "2016-04-06T08:29:27"
        self.stm = Steem(offline=True)

    def doit(self, printWire=False, ops=None):
        ops = [Operation(ops)]
        tx = Signed_Transaction(
            ref_block_num=self.ref_block_num,
            ref_block_prefix=self.ref_block_prefix,
            expiration=self.expiration,
            operations=ops,
        )
        start = timer()
        tx = tx.sign([self.wif], chain=self.prefix)
        end1 = timer()
        tx.verify([PrivateKey(self.wif, prefix="STM").pubkey], self.prefix)
        end2 = timer()
        return end2 - end1, end1 - start


class SteemTest(object):
    def setup(self):
        self.prefix = "STEEM"
        self.default_prefix = "STM"
        self.wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"
        self.ref_block_num = 34294
        self.ref_block_prefix = 3707022213
        self.expiration = "2016-04-06T08:29:27"

    def doit(self, printWire=False, ops=None):
        ops = [steemOperations.Operation(ops)]
        tx = steemSignedTransaction(
            ref_block_num=self.ref_block_num,
            ref_block_prefix=self.ref_block_prefix,
            expiration=self.expiration,
            operations=ops,
        )
        start = timer()
        tx = tx.sign([self.wif], chain=self.prefix)
        end1 = timer()
        tx.verify([steemPrivateKey(self.wif, prefix="STM").pubkey], self.prefix)
        end2 = timer()
        return end2 - end1, end1 - start


if __name__ == "__main__":
    steem_test = SteemTest()
    nectar_test = nectarTest()
    steem_test.setup()
    nectar_test.setup()
    steem_times = []
    nectar_times = []
    loops = 50
    for i in range(0, loops):
        print(i)
        opSteem = steemOperations.Transfer(
            **{"from": "foo", "to": "baar", "amount": "111.110 STEEM", "memo": "Fooo"}
        )
        opnectar = operations.Transfer(
            **{
                "from": "foo",
                "to": "baar",
                "amount": Amount("111.110 STEEM", steem_instance=Steem(offline=True)),
                "memo": "Fooo",
            }
        )

        t_s, t_v = steem_test.doit(ops=opSteem)
        steem_times.append([t_s, t_v])

        t_s, t_v = nectar_test.doit(ops=opnectar)
        nectar_times.append([t_s, t_v])

    steem_dt = [0, 0]
    nectar_dt = [0, 0]
    for i in range(0, loops):
        steem_dt[0] += steem_times[i][0]
        steem_dt[1] += steem_times[i][1]
        nectar_dt[0] += nectar_times[i][0]
        nectar_dt[1] += nectar_times[i][1]
    print("steem vs nectar:\n")
    print("steem: sign: %.2f s, verification %.2f s" % (steem_dt[0] / loops, steem_dt[1] / loops))
    print("nectar:  sign: %.2f s, verification %.2f s" % (nectar_dt[0] / loops, nectar_dt[1] / loops))
    print("------------------------------------")
    print(
        "nectar is %.2f %% (sign) and %.2f %% (verify) faster than steem"
        % (steem_dt[0] / nectar_dt[0] * 100, steem_dt[1] / nectar_dt[1] * 100)
    )



# -*- coding: utf-8 -*-
import json
import logging
from timeit import default_timer as timer

from nectar.account import Account
from nectar.instance import shared_blockchain_instance

log = logging.getLogger(__name__)


def node_answer_time(node):
    try:
        from nectar.blockchaininstance import BlockChainInstance

        stm_local = BlockChainInstance(node=node, num_retries=2, num_retries_call=2, timeout=10)
        start = timer()
        stm_local.get_network(use_stored_data=False)
        stop = timer()
        rpc_answer_time = stop - start
    except KeyboardInterrupt:
        rpc_answer_time = float("inf")
        raise KeyboardInterrupt()
    except Exception:
        rpc_answer_time = float("inf")
    return rpc_answer_time


class NodeList(list):
    """Returns HIVE/STEEM nodes as list

    .. code-block:: python

        from nectar.nodelist import NodeList
        n = NodeList()
        nodes_urls = n.get_nodes()

    """

    def __init__(self):
        nodes = [
            {
                "url": "https://api.steemit.com",
                "version": "0.20.2",
                "type": "appbase",
                "owner": "steemit",
                "hive": False,
                "score": 50,
            },
            {
                "url": "https://api.hive.blog",
                "version": "1.27.8",
                "type": "appbase",
                "owner": "hive",
                "hive": True,
                "score": 80,
            },
            {
                "url": "https://api.syncad.com",
                "version": "1.27.8",
                "type": "appbase",
                "owner": "syncad",
                "hive": True,
                "score": 70,
            },
            {
                "url": "https://anyx.io",
                "version": "1.27.4",
                "type": "appbase",
                "owner": "anyx",
                "hive": True,
                "score": 60,
            },
            {
                "url": "https://api.openhive.network",
                "version": "1.27.8",
                "type": "appbase",
                "owner": "gtg",
                "hive": True,
                "score": 50,
            },
            {
                "url": "https://rpc.mahdiyari.info",
                "version": "1.27.5",
                "type": "appbase",
                "owner": "mahdiyari",
                "hive": True,
                "score": 50,
            },
            {
                "url": "https://api.c0ff33a.uk",
                "version": "1.27.5",
                "type": "appbase",
                "owner": "c0ff33a",
                "hive": True,
                "score": 40,
            },
            {
                "url": "https://api.deathwing.me",
                "version": "1.27.5",
                "type": "appbase",
                "owner": "deathwing",
                "hive": True,
                "score": 40,
            },
            {
                "url": "https://hive-api.3speak.tv",
                "version": "1.27.5",
                "type": "appbase",
                "owner": "3speak",
                "hive": True,
                "score": 40,
            },
            {
                "url": "https://hive-api.arcange.eu",
                "version": "1.27.5",
                "type": "appbase",
                "owner": "arcange",
                "hive": True,
                "score": 40,
            },
            {
                "url": "https://hive-api.dlux.io",
                "version": "1.27.8",
                "type": "appbase",
                "owner": "dlux",
                "hive": True,
                "score": 30,
            },
            {
                "url": "https://api.hive.blue",
                "version": "1.27.5",
                "type": "appbase",
                "owner": "hive.blue",
                "hive": True,
                "score": 30,
            },
            {
                "url": "https://hiveapi.actifit.io",
                "version": "1.27.8",
                "type": "appbase",
                "owner": "actifit",
                "hive": True,
                "score": 30,
            },
            {
                "url": "https://techcoderx.com",
                "version": "1.27.7",
                "type": "appbase",
                "owner": "techcoderx",
                "hive": True,
                "score": 20,
            },
            {
                "url": "https://hive-test-beeabode.roelandp.nl",
                "version": "0.23.0",
                "type": "testnet",
                "owner": "roelandp",
                "hive": True,
                "score": 5,
            },
        ]
        super(NodeList, self).__init__(nodes)

    def update(self, node_list):
        new_nodes = []
        for node_url in node_list:
            node_found = False
            for node in self:
                if node["url"] == node_url:
                    new_nodes.append(node)
                    node_found = True
                    break
            if not node_found:
                log.warning(f"Node {node_url} not found in the original list")

        super(NodeList, self).__init__(new_nodes)

    def get_node_answer_time(self, node_list=None, verbose=False):
        """Pings all nodes and measure the answer time

        .. code-block:: python

            from nectar.nodelist import NodeList
            nl = NodeList()
            nl.update_nodes()
            nl.ping_nodes()
        """
        ping_times = []
        if node_list is None:
            node_list = []
            for node in self:
                node_list.append(node["url"])
        for node in node_list:
            ping_times.append(1000.0)
        available_nodes = []
        for node in self:
            available_nodes.append(node["url"])
        for i in range(len(node_list)):
            if node_list[i] not in available_nodes:
                ping_times[i] = float("inf")
                continue
            try:
                ping_times[i] = node_answer_time(node_list[i])
                if verbose:
                    log.info("node %s results in %.2f" % (node_list[i], ping_times[i]))
            except KeyboardInterrupt:
                ping_times[i] = float("inf")
                break
        sorted_arg = sorted(range(len(ping_times)), key=ping_times.__getitem__)
        sorted_nodes = []
        for i in sorted_arg:
            if ping_times[i] != float("inf"):
                sorted_nodes.append({"url": node_list[i], "delay_ms": ping_times[i] * 1000})
        return sorted_nodes

    def update_nodes(self, weights=None, blockchain_instance=None, **kwargs):
        """Reads metadata from nectarflower and recalculates the nodes score

        :param list/dict weight: can be used to weight the different benchmarks
        :type weight: list, dict

        .. code-block:: python

            from nectar.nodelist import NodeList
            nl = NodeList()
            weights = [0, 0.1, 0.2, 1]
            nl.update_nodes(weights)
            weights = {'block': 0.1, 'history': 0.1, 'apicall': 1, 'config': 1}
            nl.update_nodes(weights)
        """
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        steem = blockchain_instance or shared_blockchain_instance()

        metadata = None
        account = None
        cnt = 0
        while metadata is None and cnt < 5:
            cnt += 1
            try:
                account = Account("nectarflower", blockchain_instance=steem)
                # Metadata is stored in the account's json_metadata field (not posting_json_metadata)
                raw_meta = account.get("json_metadata") or ""
                try:
                    metadata = json.loads(raw_meta) if raw_meta else None
                except Exception:
                    metadata = None
            except Exception as e:
                log.warning(f"Error fetching metadata (attempt {cnt}): {str(e)}")
                steem.rpc.next()
                account = None
                metadata = None

        if metadata is None:
            log.warning("Failed to fetch nectarflower metadata after multiple attempts")
            return

        report = metadata.get("report", [])
        failing_nodes = metadata.get("failing_nodes", {})
        parameter = metadata.get("parameter", {})
        benchmarks = parameter.get("benchmarks")

        # Determine benchmark names. If not explicitly provided in metadata parameters, derive them
        # by inspecting the keys of the report entries and filtering out the non-benchmark fields.
        if benchmarks and isinstance(benchmarks, dict):
            benchmark_names: list[str] = list(benchmarks.keys())
        else:
            benchmark_names = []
            # Common non-benchmark keys present in every report entry
            _skip_keys = {
                "node",
                "version",
                "hive",
                "weighted_score",
                "tests_completed",
            }
            # Collect benchmark names dynamically from the report section
            for _entry in report:
                if isinstance(_entry, dict):
                    for _k in _entry.keys():
                        if _k not in _skip_keys and _k not in benchmark_names:
                            benchmark_names.append(_k)
            # Sort for deterministic ordering
            benchmark_names.sort()

        if weights is None:
            weights_dict = {}
            for benchmark in benchmark_names:
                weights_dict[benchmark] = 1.0 / len(benchmark_names)
        elif isinstance(weights, list):
            weights_dict = {}
            i = 0
            weight_sum = sum(weights)
            for benchmark in benchmark_names:
                if i < len(weights):
                    weights_dict[benchmark] = weights[i] / weight_sum if weight_sum > 0 else 0
                else:
                    weights_dict[benchmark] = 0.0
                i += 1
        elif isinstance(weights, dict):
            weights_dict = {}
            weight_sum = sum(weights.values())
            for benchmark in benchmark_names:
                if benchmark in weights:
                    weights_dict[benchmark] = (
                        weights[benchmark] / weight_sum if weight_sum > 0 else 0
                    )
                else:
                    weights_dict[benchmark] = 0.0

        max_score = len(report) + 1
        new_nodes = []
        update_count = 0
        failing_count = 0

        for node in self:
            new_node = node.copy()
            node_was_updated = False

            # Check against report data
            for report_node in report:
                if node["url"] == report_node.get("node", ""):
                    update_count += 1
                    new_node["version"] = report_node.get(
                        "version", new_node.get("version", "0.0.0")
                    )
                    new_node["hive"] = report_node.get("hive", new_node.get("hive", False))

                    scores = []
                    for benchmark in benchmark_names:
                        if benchmark in report_node:
                            result = report_node[benchmark]
                            rank = result.get("rank", -1)
                            if not result.get("ok", False):
                                rank = max_score + 1
                            score = (
                                (max_score - rank) / (max_score - 1) * 100
                                if rank > 0 and max_score > 1
                                else 0
                            )
                            weighted_score = score * weights_dict.get(benchmark, 0)
                            scores.append(weighted_score)

                    # Prefer the pre-computed weighted_score from the metadata if present; fall back to
                    # the locally calculated score otherwise.
                    if "weighted_score" in report_node and isinstance(
                        report_node["weighted_score"], (int, float)
                    ):
                        new_node["score"] = report_node["weighted_score"]
                    else:
                        sum_score = sum(scores)
                        new_node["score"] = sum_score
                    node_was_updated = True
                    break

            # Check if node is in failing nodes list
            if node["url"] in failing_nodes:
                failing_count += 1
                new_node["score"] = -1
            elif not node_was_updated:
                # If node wasn't part of the metadata report, reset its score so it
                # doesn't overshadow the authoritative ordering
                new_node["score"] = 0

            new_nodes.append(new_node)

        # ------------------------------------------------------------
        # Ensure that all nodes present in the metadata are included
        # in the final list, even if they were not part of the default
        # hard-coded set.
        # ------------------------------------------------------------
        existing_urls: set[str] = {n["url"] for n in new_nodes}

        # Add nodes found in the report section
        for report_node in report:
            url = report_node.get("node")
            if not url or url in existing_urls:
                continue
            new_entry = {
                "url": url,
                "version": report_node.get("version", "0.0.0"),
                "type": "appbase",
                "owner": report_node.get("owner", "unknown"),
                "hive": report_node.get("hive", True),
                "score": report_node.get("weighted_score", 0),
            }
            new_nodes.append(new_entry)
            existing_urls.add(url)

        # Add nodes listed as failing but missing
        for url in failing_nodes.keys():
            if url in existing_urls:
                continue
            new_nodes.append(
                {
                    "url": url,
                    "version": "unknown",
                    "type": "appbase",
                    "owner": "unknown",
                    "hive": True,
                    "score": -1,
                }
            )
            existing_urls.add(url)

        # Re-initialise internal list
        super(NodeList, self).__init__(new_nodes)

    def get_nodes(
        self,
        hive=False,
        exclude_limited=False,
        dev=False,
        testnet=False,
        testnetdev=False,
        wss=True,
        https=True,
        not_working=False,
        normal=True,
        appbase=True,
    ):
        """Returns nodes as list

        :param bool hive: When True, only HIVE nodes will be returned
        :param bool exclude_limited: When True, limited nodes are excluded
        :param bool dev: when True, dev nodes with version 0.19.11 are included
        :param bool testnet: when True, testnet nodes are included
        :param bool testnetdev: When True, testnet-dev nodes are included
        :param bool not_working: When True, all nodes including not working ones will be returned
        :param bool normal: deprecated
        :param bool appbase: deprecated

        """
        node_list = []
        node_type_list = []
        if normal:
            node_type_list.append("normal")
        if appbase:
            node_type_list.append("appbase")
        if dev:
            node_type_list.append("appbase-dev")
        if testnet:
            node_type_list.append("testnet")
        if testnetdev:
            node_type_list.append("testnet-dev")
        if not exclude_limited:
            node_type_list.append("appbase-limited")
        for node in self:
            if node["type"] in node_type_list and (node["score"] >= 0 or not_working):
                if hive != node["hive"]:
                    continue
                if not https and node["url"][:5] == "https":
                    continue
                if not wss and node["url"][:3] == "wss":
                    continue
                node_list.append(node)

        return [
            node["url"] for node in sorted(node_list, key=lambda self: self["score"], reverse=True)
        ]

    def get_hive_nodes(self, testnet=False, not_working=False, wss=True, https=True):
        """Returns hive only nodes as list

        :param bool testnet: when True, testnet nodes are included
        :param bool not_working: When True, all nodes including not working ones will be returned

        """
        node_list = []

        for node in self:
            if not node["hive"]:
                continue
            if node["score"] < 0 and not not_working:
                continue
            if (testnet and node["type"] == "testnet") or (
                not testnet and node["type"] != "testnet"
            ):
                if not https and node["url"][:5] == "https":
                    continue
                if not wss and node["url"][:3] == "wss":
                    continue
                node_list.append(node)

        return [
            node["url"] for node in sorted(node_list, key=lambda self: self["score"], reverse=True)
        ]

    def get_steem_nodes(self, testnet=False, not_working=False, wss=True, https=True):
        """Returns steem only nodes as list

        :param bool testnet: when True, testnet nodes are included
        :param bool not_working: When True, all nodes including not working ones will be returned

        """
        node_list = []

        for node in self:
            if node["hive"]:
                continue
            if node["score"] < 0 and not not_working:
                continue
            if (testnet and node["type"] == "testnet") or (
                not testnet and node["type"] != "testnet"
            ):
                if not https and node["url"][:5] == "https":
                    continue
                if not wss and node["url"][:3] == "wss":
                    continue
                node_list.append(node)

        return [
            node["url"] for node in sorted(node_list, key=lambda self: self["score"], reverse=True)
        ]

    def get_testnet(self, testnet=True, testnetdev=False):
        """Returns testnet nodes"""
        return self.get_nodes(normal=False, appbase=False, testnet=testnet, testnetdev=testnetdev)



# -*- coding: utf-8 -*-
# Inspired by https://raw.githubusercontent.com/xeroc/python-graphenelib/master/graphenestorage/masterpassword.py
import hashlib
import logging
import os
import warnings
from binascii import hexlify

from nectargraphenebase import bip38
from nectargraphenebase.aes import AESCipher

from .exceptions import WalletLocked, WrongMasterPasswordException

log = logging.getLogger(__name__)


class MasterPassword(object):
    """The keys are encrypted with a Masterpassword that is stored in
    the configurationStore. It has a checksum to verify correctness
    of the password
    The encrypted private keys in `keys` are encrypted with a random
    **masterkey/masterpassword** that is stored in the configuration
    encrypted by the user-provided password.

    :param ConfigStore config: Configuration store to get access to the
        encrypted master password
    """

    def __init__(self, config=None, **kwargs):
        if config is None:
            raise ValueError("If using encrypted store, a config store is required!")
        self.config = config
        self.password = None
        self.decrypted_master = None
        self.config_key = "encrypted_master_password"

    @property
    def masterkey(self):
        """Contains the **decrypted** master key"""
        return self.decrypted_master

    def has_masterpassword(self):
        """Tells us if the config store knows an encrypted masterpassword"""
        return self.config_key in self.config

    def locked(self):
        """Is the store locked. E.g. Is a valid password known that can be
        used to decrypt the master key?
        """
        return not self.unlocked()

    def unlocked(self):
        """Is the store unlocked so that I can decrypt the content?"""
        if self.password is not None:
            return bool(self.password)
        else:
            password_storage = self.config["password_storage"]
            KEYRING_AVAILABLE = False
            if password_storage == "keyring":
                try:
                    import keyring

                    if not isinstance(keyring.get_keyring(), keyring.backends.fail.Keyring):
                        KEYRING_AVAILABLE = True
                    else:
                        KEYRING_AVAILABLE = False
                except ImportError:
                    KEYRING_AVAILABLE = False
            if (
                "UNLOCK" in os.environ
                and os.environ["UNLOCK"]
                and self.config_key in self.config
                and self.config[self.config_key]
                and password_storage == "environment"
            ):
                log.debug("Trying to use environmental variable to unlock wallet")
                self.unlock(os.environ.get("UNLOCK"))
                return bool(self.password)
            elif (
                password_storage == "keyring"
                and KEYRING_AVAILABLE
                and self.config_key in self.config
                and self.config[self.config_key]
            ):
                log.debug("Trying to use keyring to unlock wallet")
                pwd = keyring.get_password("nectar", "wallet")
                self.unlock(pwd)
                return bool(self.password)
        return False

    def lock(self):
        """Lock the store so that we can no longer decrypt the content of the
        store
        """
        self.password = None
        self.decrypted_master = None

    def unlock(self, password):
        """The password is used to encrypt this masterpassword. To
        decrypt the keys stored in the keys database, one must use
        BIP38, decrypt the masterpassword from the configuration
        store with the user password, and use the decrypted
        masterpassword to decrypt the BIP38 encrypted private keys
        from the keys storage!

        :param str password: Password to use for en-/de-cryption
        """
        self.password = password
        if self.config_key in self.config and self.config[self.config_key]:
            self._decrypt_masterpassword()
        else:
            self._new_masterpassword(password)
            self._save_encrypted_masterpassword()

    def wipe_masterpassword(self):
        """Removes the encrypted masterpassword from config storage"""
        if self.config_key in self.config and self.config[self.config_key]:
            self.config[self.config_key] = None

    def _decrypt_masterpassword(self):
        """Decrypt the encrypted masterkey"""
        aes = AESCipher(self.password)
        checksum, encrypted_master = self.config[self.config_key].split("$")
        try:
            decrypted_master = aes.decrypt(encrypted_master)
        except Exception:
            self._raise_wrongmasterpassexception()
        if checksum != self._derive_checksum(decrypted_master):
            self._raise_wrongmasterpassexception()
        self.decrypted_master = decrypted_master

    def _raise_wrongmasterpassexception(self):
        self.password = None
        raise WrongMasterPasswordException

    def _save_encrypted_masterpassword(self):
        self.config[self.config_key] = self._get_encrypted_masterpassword()

    def _new_masterpassword(self, password):
        """Generate a new random masterkey, encrypt it with the password and
        store it in the store.

        :param str password: Password to use for en-/de-cryption
        """
        # make sure to not overwrite an existing key
        if self.config_key in self.config and self.config[self.config_key]:
            raise Exception("Storage already has a masterpassword!")

        self.decrypted_master = hexlify(os.urandom(32)).decode("ascii")

        # Encrypt and save master
        self.password = password
        self._save_encrypted_masterpassword()
        return self.masterkey

    def _derive_checksum(self, s):
        """Derive the checksum

        :param str s: Random string for which to derive the checksum
        """
        checksum = hashlib.sha256(bytes(s, "ascii")).hexdigest()
        return checksum[:4]

    def _get_encrypted_masterpassword(self):
        """Obtain the encrypted masterkey

        .. note:: The encrypted masterkey is checksummed, so that we can
            figure out that a provided password is correct or not. The
            checksum is only 4 bytes long!
        """
        if not self.unlocked():
            raise WalletLocked
        aes = AESCipher(self.password)
        return "{}${}".format(self._derive_checksum(self.masterkey), aes.encrypt(self.masterkey))

    def change_password(self, newpassword):
        """Change the password that allows to decrypt the master key"""
        if not self.unlocked():
            raise WalletLocked
        self.password = newpassword
        self._save_encrypted_masterpassword()

    def decrypt(self, wif):
        """Decrypt the content according to BIP38

        :param str wif: Encrypted key
        """
        if not self.unlocked():
            raise WalletLocked
        return format(bip38.decrypt(wif, self.masterkey), "wif")

    def deriveChecksum(self, s):
        """Derive the checksum"""
        checksum = hashlib.sha256(bytes(s, "ascii")).hexdigest()
        return checksum[:4]

    def encrypt_text(self, txt):
        """Encrypt the content according to BIP38

        :param str wif: Unencrypted key
        """
        if not self.unlocked():
            raise WalletLocked
        aes = AESCipher(self.masterkey)
        return "{}${}".format(self.deriveChecksum(txt), aes.encrypt(txt))

    def decrypt_text(self, enctxt):
        """Decrypt the content according to BIP38

        :param str wif: Encrypted key
        """
        if not self.unlocked():
            raise WalletLocked
        aes = AESCipher(self.masterkey)
        checksum, encrypted_text = enctxt.split("$")
        try:
            decrypted_text = aes.decrypt(encrypted_text)
        except Exception:
            raise WrongMasterPasswordException
        if checksum != self.deriveChecksum(decrypted_text):
            raise WrongMasterPasswordException
        return decrypted_text

    def encrypt(self, wif):
        """Encrypt the content according to BIP38

        :param str wif: Unencrypted key
        """
        if not self.unlocked():
            raise WalletLocked
        return format(bip38.encrypt(str(wif), self.masterkey), "encwif")

    def changePassword(self, newpassword):  # pragma: no cover
        warnings.warn(
            "changePassword will be replaced by change_password in the future",
            PendingDeprecationWarning,
        )
        return self.change_password(newpassword)
